/* $Id: interp.h,v 1.27 2002/12/09 10:35:20 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifndef INTERP_H
#define INTERP_H

#define SCRIPT_PATH_ENV "SCRIPT_PATH"
#define SL_TABLE_PATH_ENV "SL_TABLE_PATH"
#define SL_TABLE_NEW_PATH_ENV "SL_TABLE_NEW_PATH"
#define SL_TABLE_RO_PATH_ENV "SL_TABLE_RO_PATH"

#include <limits.h>
#include "lists.h"
#include "dstring.h"
#include "shared.h"
#include "automat.h"
#include "parsehtml.h"
#include "flow.h"

struct procedure;

DEFLIST(trap_handler,
		char *trapname;
		struct procedure *proc;
		);

struct interpreter;
struct chunk;
struct variable;
struct argument;

DEFLIST(value, dstring *val;);

#define next_value(val) ((value *)list_next(val))
#define third_value(val) ((value *)list_next(list_next(val)))

typedef void (*instruction_t)(struct chunk *, struct interpreter *,  value *);

enum chunk_types { arg_literal, arg_variable, arg_parameter, 
 				   arg_funcall, arg_builtin, arg_embedded, 
				   arg_statement
				  };

enum special_parameters { param_count = -1, switch_header = -1024};

DEFLIST(argument, struct chunk *arg;);	   

DEFLIST(branch,
		argument *condition;
		struct chunk *loc;
		);

typedef struct ctl_data
{
	argument *header;
	struct chunk *ctl_end;
	struct chunk *on_break;
	branch *branches;
} ctl_data;

DEFLIST(chunk,
		char *file;
		int lno;
        int type;
        union
		{
			dstring *literal;
			struct
			{
				unsigned char is_tag : 1;
				char *name;
			} var;
			int paramno;
			struct
			{
				union
				{
					char *name;
					instruction_t bi;
					struct chunk *emb;
				} f;
				struct argument *args;
			} func;
			struct
			{
				int statement;
				ctl_data *ctl;
			} s;
        } x;
       );


enum variable_status { var_local, var_global, var_shared };
enum statement_types { s_if, s_foreach, s_while, s_switch };

struct storage;	
DEFLIST(ctlstack,
		 chunk *master;
		 branch *cur_branch;
		 union 
		 {
		 	struct loopstack
			{
				 struct storage *hangon;
				 void *param;
				 struct variable *var;
				 unsigned char first_time : 1;
			} loop;
			dstring *condition; 
		  } x;
       );


DEFLIST(variable,
		shared_object sh;
        value *vals;
		unsigned char readonly : 1;
       );

DEFLIST(execstack, 
		chunk *curfunc;
		chunk *curop;
		value *out_vals;
		value *last_val;
		argument *nextarg;
		dstring *buffer;
		);

struct shared_pool;

DEFLIST(local_storage,
		shared_int *sto;
	);

DEFLIST(procstack,
		 struct interpreter *where;
         ctlstack *ctl_stack;
		 chunk *nexttop;
		 value *args;
		 shared_int *vars;
		 execstack *state;
		 struct shared_pool *cur_pool;
		 local_storage *local_stos;
		 int trace_flag;
		 int boundary;
       );

DEFLIST(procedure,
		shared_object sh;
		chunk *body; 
		unsigned char callback : 1;
		struct shared_pool *pool;
	);

typedef struct master_instruction_src
{
	char *name;
	int returns_value;
	instruction_t action;
	int minarg;
	int maxarg;
	int defarg;
	int optimizable;
} master_instruction_src;

DEFLIST(declaration,
	char *name;
	int returns_value;
	int minarg;
	int maxarg;
	int defarg;
	int optimizable;
);
	
typedef struct instruction_src
{
	char *name;
    instruction_t action;
} instruction_src;

typedef struct storage_class
{
	struct storage *(*create)(struct interpreter *i, char *name, char *arg);
	void (*destroy)(struct interpreter *i, struct storage *s);
	int (*check)(struct interpreter *i, struct storage *s, dstring *key);
	dstring *(*get)(struct interpreter *i, struct storage *s, dstring *key);
	void (*put)(struct interpreter *i, struct storage *s, dstring *key, dstring *val);
	void (*add)(struct interpreter *i, struct storage *s, dstring *key, dstring *val);
	void (*remove)(struct interpreter *i, struct storage *s, dstring *key);
	dstring *(*first)(struct interpreter *i, struct storage *s, void **iter);
	dstring *(*next)(struct interpreter *i, struct storage *s, void **iter);
	void (*enditer)(struct interpreter *i, struct storage *s, void **iter);
	int (*numofitems)(struct interpreter *i, struct storage *s);
	dstring *(*seek)(struct interpreter *i, struct storage *s, void **iter, dstring *key); 
	struct storage *(*clone)(struct storage *src);
} storage_class;


DEFLIST(to_remove_list,
		dstring *key;
		);

DEFLIST(storage, 
		shared_object sh;
		storage_class *class;
		void *data;
		to_remove_list *to_remove;
);

struct stream;
typedef struct stream_class
{
	char *name;
	struct stream *(*open)(struct interpreter *i, char *name, char *arg, int writeable);
	void (*close)(struct interpreter *i, struct stream *s);
	html_chunk *(*read_chunk)(struct interpreter *i, struct stream *s);
	void (*writestr)(struct interpreter *i, struct stream *s, char *str);
	void (*writetag)(struct interpreter *i, struct stream *s, html_chunk *chunk);
	char *(*ioctl)(struct interpreter *i, struct stream *s, char *ctlstr);
	int (*getpos)(struct interpreter *i, struct stream *s);
	void (*setpos)(struct interpreter *i, struct stream *s, int pos);
	void (*rewind)(struct interpreter *i, struct stream *s);
} stream_class;


enum stream_directions { stream_in, stream_out };

DEFLIST(stream, 
		shared_object sh;
		stream_class *class;
		int direction;
		void *data;
);

DEFLIST(mod_stream,
		stream_class *cls;
		);

typedef struct automaton_ext
{
	struct interpreter *env;
	struct procedure *process;
	struct procedure *destroy;
	struct procedure *reset;
	struct procedure *isfinal;
	struct procedure *getreg;
	struct procedure *mapname;
} automaton_ext;

DEFLIST(autoseqdef,
		shared_object sh;
		autoseq *seq;
	   );

DEFLIST(shared_pool,
		int refcnt;
		char *name;
		shared_int *vars;
);

DEFLIST(tracelist,
		char *pattern;
		);


DEFLIST(interpreter,
	shared_object sh;
    chunk *main;
	chunk *atexit;
    shared_int *procs;
    shared_int *tags;
	shared_int *filters;
	shared_int *char_cache[UCHAR_MAX + 1];
	instruction_src *builtins;
	declaration *decls;
    shared_int *globvars;
	html_chunk *this_chunk;
	char *prefix;
	shared_int *aseqs;
	shared_int *stos;
	shared_int *streams;
	shared_int *subints;
	stream *instream;
	stream *outstream;
	tracelist *to_trace;
	int line_tracing;
	int prev_line;
	procedure *error_handler;
	mod_stream *mod_streams;
	struct interpreter *forwarder;
);

int convert_to_int(char *str);

value *copy_values(value *vals);

variable *safe_get_var(interpreter *i, char *name, int writing);
variable *get_shared_value(char *poolname, char *name);

#define STORAGE_CLASS "\1"
#define STREAM_CLASS "\2"
#define AUTOSEQ_CLASS "\3"
#define PROC_CLASS "\4"
#define SUBINT_CLASS "\5"
#define VAR_CLASS "\6"
/* Special classes */
#define LOCALVAR_CLASS "\10"
#define POOLVAR_CLASS "\11"
#define TAG_CLASS "\12"
#define FILTER_CLASS "\13"

#define init_shared_stream(i, so, intname) init_shared_object((i), (so), STREAM_CLASS, (intname))
#define init_shared_storage(i, so, intname) init_shared_object((i), (so), STORAGE_CLASS, (intname))
#define init_shared_autoseq(i, so, intname) init_shared_object((i), (so), AUTOSEQ_CLASS, (intname))
#define init_shared_proc(i, so, intname) init_shared_object((i), (so), PROC_CLASS, (intname))
#define init_shared_subint(i, so, intname) init_shared_object((i), (so), SUBINT_CLASS, (intname))
#define init_shared_var(env, so, kind, intname) init_shared_object((env), (so), kind, (intname))

#define find_storage(i, name) ((storage *)find_shared_int((i), STORAGE_CLASS, (name)))
#define find_stream(i, name) ((stream *)find_shared_int((i), STREAM_CLASS, (name)))
#define find_asd(i, name) ((autoseqdef *)find_shared_int((i), AUTOSEQ_CLASS, (name)))
#define find_proc(i, name) ((procedure *)find_shared_int((i), PROC_CLASS, (name)))
#define qfind_proc(i, name) ((procedure *)quiet_find_shared_int((i), PROC_CLASS, (name)))
#define find_subint(i, name) ((interpreter *)find_shared_int((i), SUBINT_CLASS, (name)))
#define qfind_var(env, kind, name) ((variable *)quiet_find_shared_int((env), kind, (name)))

storage *create_storagex(interpreter *i, storage_class *class, char **name, char *arg);
storage *create_storage(interpreter *i, storage_class *class, char *name, char *arg);
void register_mod_stream(interpreter *i, stream_class *cls);
stream_class *detect_stream_class(interpreter *i, char *id, dstring **options);

void discard_value(dstring *ds);
	
procstack *get_pstack(void);
dstring *eval_procedure(interpreter *i, char *procname, value *args);
dstring *eval_procedurev(interpreter *i, char *procname, ...);
dstring *eval_procedurepv(interpreter *i, procedure *proc, ...);
interpreter *load_interpreterx(char *fname, interpreter *parent, stream *ins, stream *outs);
#define load_interpreter(fname, ins, outs) (load_interpreterx(fname, NULL, ins, outs))
void reinit_interpreter(interpreter *i, stream *ins, stream *outs);
void cleanup_interpreter(interpreter *i);

instruction_src *lookup_instruction(interpreter *i, char *name);

void add_declaration(interpreter *i, declaration *decl);
void override_declaration(interpreter *i, declaration *decl);
void add_declaration_master(interpreter *i, master_instruction_src *ms);
void parse_declaration(interpreter *i, char *name, int override, char *dclstr, char **end);
dstring *stringize_declaration(declaration *decl);
declaration *lookup_declaration(interpreter *i, char *name);

void merge_builtins(interpreter *i,  master_instruction_src *add);

int script_main_loop(interpreter *i);


#endif
